from django.core.validators import MinValueValidator
from django.db import models
from django.db.models.deletion import PROTECT
from edc_model import models as edc_models
from edc_sites.models import CurrentSiteManager, SiteModelMixin
from edc_utils.date import get_utcnow_as_date
from edc_visit_schedule.model_mixins import VisitCodeFieldsModelMixin

from ..dosage_per_day import dosage_per_day
from ..exceptions import ActivePrescriptionRefillExists
from .dosage_guideline import DosageGuideline
from .formulation import Formulation
from .list_models import FrequencyUnits
from .model_mixins import MedicationOrderModelMixin
from .rx import Rx


class Manager(models.Manager):

    use_in_migrations = True

    def get_by_natural_key(self, prescription, medication, refill_date):
        return self.get(prescription, medication, refill_date)


class RxRefill(
    MedicationOrderModelMixin,
    VisitCodeFieldsModelMixin,
    SiteModelMixin,
    edc_models.BaseUuidModel,
):

    rx = models.ForeignKey(Rx, on_delete=PROTECT)

    dosage_guideline = models.ForeignKey(DosageGuideline, on_delete=PROTECT)

    formulation = models.ForeignKey(Formulation, on_delete=PROTECT, null=True)

    dose = models.DecimalField(
        max_digits=6,
        decimal_places=1,
        null=True,
        blank=True,
        help_text="dose per frequency if NOT considering weight",
    )

    calculate_dose = models.BooleanField(default=True)

    frequency = models.IntegerField(
        validators=[MinValueValidator(1)],
        null=True,
        blank=True,
    )

    frequency_units = models.ForeignKey(
        FrequencyUnits,
        verbose_name="per",
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    weight_in_kgs = models.DecimalField(
        max_digits=6, decimal_places=1, null=True, blank=True
    )

    refill_date = models.DateField(
        verbose_name="Refill date", default=get_utcnow_as_date, help_text=""
    )

    number_of_days = models.IntegerField(null=True)

    total = models.DecimalField(
        max_digits=6,
        decimal_places=1,
        null=True,
        blank=True,
        help_text="Leave blank to auto-calculate",
    )

    remaining = models.DecimalField(
        max_digits=6,
        decimal_places=1,
        null=True,
        blank=True,
        help_text="Leave blank to auto-calculate",
    )

    notes = models.TextField(
        max_length=250,
        null=True,
        blank=True,
        help_text="Additional information for patient",
    )

    active = models.BooleanField(default=False)

    verified = models.BooleanField(default=False)
    verified_datetime = models.DateTimeField(null=True, blank=True)

    as_string = models.CharField(max_length=150, editable=False)

    on_site = CurrentSiteManager()

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return (
            f"{self.rx} "
            f"Take {self.dose} {self.formulation.formulation_type.display_name} {self.formulation.route.display_name} "
            # f"{self.frequency} {self.frequency_units.display_name}"
        )

    def natural_key(self):
        return (
            self.rx,
            self.medication,
            self.refill_date,
        )

    def save(self, *args, **kwargs):
        if self.active:
            # TODO: does this matter?
            opts = dict(id=self.id) if self.id else {}
            if (
                self.__class__.objects.filter(
                    rx__subject_identifier=self.rx.subject_identifier,
                    dosage_guideline=self.dosage_guideline,
                    active=True,
                )
                .exclude(**opts)
                .exists()
            ):
                raise ActivePrescriptionRefillExists(
                    f"Unable to save as an active refill. An active refill already exists."
                )

        self.medication = self.dosage_guideline.medication
        # if not self.dose and self.calculate_dose:
        self.frequency = self.dosage_guideline.frequency
        self.frequency_units = self.dosage_guideline.frequency_units
        self.dose = self.get_dose()
        self.total = self.get_total()
        if not self.id:
            self.remaining = self.total
        self.as_string = str(self)
        super().save(*args, **kwargs)

    def get_dose(self):
        return dosage_per_day(
            self.dosage_guideline,
            weight_in_kgs=self.weight_in_kgs,
            strength=self.formulation.strength,
            strength_units=self.formulation.units.name,
        )

    def get_total(self):
        return float(self.dose) * float(self.number_of_days)

    @property
    def subject_identifier(self):
        return self.rx.subject_identifier

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "RX refill"
        verbose_name_plural = "RX refills"
        unique_together = [
            ["rx", "dosage_guideline", "refill_date"],
            ["rx", "visit_code", "visit_code_sequence"],
        ]



from .dispensing_history_form import (
    DispensingHistoryForm,
    DispensingHistoryReadonlyForm,
)
from .dosage_guideline_form import DosageGuidelineForm
from .formulation_form import FormulationForm
from .labels_form import LabelsForm
from .medication_form import MedicationForm
from .medication_lot_form import MedicationLotForm
from .medication_order_form import MedicationOrderForm
from .medication_product_form import MedicationProductForm
from .medication_stock_create_labels_form import MedicationStockCreateLabelsForm
from .medication_stock_form import MedicationStockForm
from .medication_stock_receiving_form import MedicationStockReceivingForm
from .return_history_form import ReturnHistoryForm
from .rx_form import RxForm
from .rx_refill_form import RxRefillForm



from django.test import TestCase, tag
from edc_pharmacy.dispensing import DispenseError
from edc_pharmacy.models import (
    DispensingHistory,
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    Route,
    Rx,
    RxRefill,
    Units,
)
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow


class TestDispense(TestCase):
    def setUp(self):
        self.subject_identifier = "12345"

        RegisteredSubject.objects.create(subject_identifier="12345")

        self.medication = Medication.objects.create(
            name="flucytosine",
            display_name="Flucytosine",
        )

        self.formulation = Formulation.objects.create(
            medication=self.medication,
            strength=500,
            units=Units.objects.get(name="mg"),
            route=Route.objects.get(display_name="Oral"),
            formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
        )

        self.dosage_guideline = DosageGuideline.objects.create(
            medication=self.medication,
            dose_per_kg=100,
            dose_units=Units.objects.get(name="mg"),
            frequency=1,
            frequency_units=FrequencyUnits.objects.get(name="day"),
        )

        self.rx = Rx.objects.create(
            subject_identifier=self.subject_identifier,
            weight_in_kgs=40,
            report_datetime=get_utcnow(),
            medication=self.medication,
        )

    @tag("disp")
    def test_dispense(self):
        rx_refill = RxRefill.objects.create(
            rx=self.rx,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            frequency=1,
            dose=None,
            refill_date=get_utcnow(),
            number_of_days=7,
        )
        from pprint import pprint

        pprint(rx_refill.__dict__)
        obj = DispensingHistory.objects.create(
            rx_refill=rx_refill,
            dispensed=8,
        )
        self.assertEqual(obj.dispensed, 8)
        rx_refill = RxRefill.objects.get(id=rx_refill.id)
        self.assertEqual(rx_refill.remaining, 56 - 8)

    def test_dispense_many(self):
        rx_refill = RxRefill.objects.create(
            rx=self.rx,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            frequency=1,
            dose=None,
            refill_date=get_utcnow(),
            number_of_days=7,
        )
        dispensed = 0
        for amount in [8, 8, 8]:
            dispensed += amount
            obj = DispensingHistory.objects.create(
                rx_refill=rx_refill,
                dispensed=8,
            )
            self.assertEqual(obj.dispensed, 8)
            rx_refill = RxRefill.objects.get(id=rx_refill.id)
            self.assertEqual(rx_refill.remaining, 56 - dispensed)

    @tag("1")
    def test_attempt_to_over_dispense(self):
        rx_refill = RxRefill.objects.create(
            rx=self.rx,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            frequency=1,
            dose=None,
            refill_date=get_utcnow(),
            number_of_days=7,
        )
        dispensed = 0
        for amount in [8, 8, 8, 8, 8, 8, 8]:
            dispensed += amount
            obj = DispensingHistory.objects.create(
                rx_refill=rx_refill,
                dispensed=8,
            )
            self.assertEqual(obj.dispensed, 8)
            rx_refill = RxRefill.objects.get(id=rx_refill.id)
            self.assertEqual(rx_refill.remaining, 48 - dispensed)
        rx_refill = RxRefill.objects.get(id=rx_refill.id)
        self.assertEqual(rx_refill.remaining, 0)
        self.assertRaises(
            DispenseError,
            DispensingHistory.objects.create,
            rx_refill=rx_refill,
            dispensed=8,
        )



from django.db.models.signals import pre_save
from django.test import TestCase, tag
from edc_appointment.creators import AppointmentsCreator
from edc_appointment.models import Appointment
from edc_constants.constants import NO, YES
from edc_facility import import_holidays
from edc_pharmacy.exceptions import (
    ActivePrescriptionRefillExists,
    NextPrescriptionRefillError,
)
from edc_pharmacy.models import (
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    MedicationOrder,
    Route,
    Rx,
    RxRefill,
    Units,
)
from edc_pharmacy.tests.models import StudyMedication, SubjectVisit
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow
from edc_visit_schedule import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED

from ..visit_schedule import schedule, visit_schedule


class TestMedicationCrf(TestCase):
    @classmethod
    def setUpTestData(cls):
        import_holidays()
        pre_save.disconnect(dispatch_uid="requires_consent_on_pre_save")

    def setUp(self) -> None:
        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False

        site_visit_schedules.register(visit_schedule)
        self.subject_identifier = "12345"
        RegisteredSubject.objects.create(subject_identifier=self.subject_identifier)
        report_datetime = get_utcnow()
        creator = AppointmentsCreator(
            subject_identifier=self.subject_identifier,
            visit_schedule=visit_schedule,
            schedule=schedule,
            report_datetime=report_datetime,
        )
        creator.create_appointments(base_appt_datetime=report_datetime)

        self.assertGreater(
            Appointment.objects.filter(
                subject_identifier=self.subject_identifier
            ).count(),
            0,
        )

        self.medication = Medication.objects.create(
            name="Flucytosine",
        )

        self.formulation = Formulation.objects.create(
            medication=self.medication,
            strength=500,
            units=Units.objects.get(name="mg"),
            route=Route.objects.get(display_name="Oral"),
            formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
        )

        self.dosage_guideline_100 = DosageGuideline.objects.create(
            medication=self.medication,
            dose_per_kg=100,
            dose_units=Units.objects.get(name="mg"),
            frequency=1,
            frequency_units=FrequencyUnits.objects.get(name="day"),
        )

        self.dosage_guideline_200 = DosageGuideline.objects.create(
            medication=self.medication,
            dose_per_kg=100,
            dose_units=Units.objects.get(name="mg"),
            frequency=2,
            frequency_units=FrequencyUnits.objects.get(name="day"),
        )

        self.rx = Rx.objects.create(
            subject_identifier=self.subject_identifier,
            weight_in_kgs=40,
            report_datetime=report_datetime,
            medication=self.medication,
        )

    @tag("35")
    def test_ok(self):
        appointment = Appointment.objects.all().order_by("timepoint")[0]
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        obj = StudyMedication.objects.create(
            subject_visit=subject_visit,
            dosage_guideline=self.dosage_guideline_100,
            formulation=self.formulation,
            order_next=YES,
            next_dosage_guideline=self.dosage_guideline_200,
            next_formulation=self.formulation,
        )
        self.assertIsNotNone(obj.number_of_days)
        self.assertEqual(obj.number_of_days, 7)

    @tag("35")
    def test_for_all_appts(self):
        """Assert for all appointments.

        Captures exception at last appointment where "next" is none
        """
        for appointment in Appointment.objects.all().order_by("timepoint"):
            subject_visit = SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=appointment.appt_datetime,
                reason=SCHEDULED,
            )
            if not appointment.next:
                self.assertRaises(
                    NextPrescriptionRefillError,
                    StudyMedication.objects.create,
                    subject_visit=subject_visit,
                    dosage_guideline=self.dosage_guideline_100,
                    formulation=self.formulation,
                    order_next=YES,
                    next_dosage_guideline=self.dosage_guideline_100,
                    next_formulation=self.formulation,
                )
                StudyMedication.objects.create(
                    subject_visit=subject_visit,
                    dosage_guideline=self.dosage_guideline_100,
                    formulation=self.formulation,
                    next_dosage_guideline=None,
                    next_formulation=None,
                    order_next=NO,
                )
            else:
                StudyMedication.objects.create(
                    subject_visit=subject_visit,
                    dosage_guideline=self.dosage_guideline_100,
                    formulation=self.formulation,
                    order_next=YES,
                    next_dosage_guideline=self.dosage_guideline_100,
                    next_formulation=self.formulation,
                )

    @tag("35")
    def test_refill_creates_next_refill(self):
        appointment = Appointment.objects.all().order_by("timepoint")[0]
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        self.assertEqual(RxRefill.objects.all().count(), 0)
        StudyMedication.objects.create(
            subject_visit=subject_visit,
            dosage_guideline=self.dosage_guideline_100,
            formulation=self.formulation,
            order_next=YES,
            next_dosage_guideline=self.dosage_guideline_200,
            next_formulation=self.formulation,
        )
        self.assertEqual(RxRefill.objects.all().count(), 2)

    @tag("35")
    def test_refill_creates_next_refill_for_next_dosage(self):
        appointment = Appointment.objects.all().order_by("timepoint")[0]
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        StudyMedication.objects.create(
            subject_visit=subject_visit,
            dosage_guideline=self.dosage_guideline_100,
            formulation=self.formulation,
            order_next=YES,
            next_dosage_guideline=self.dosage_guideline_200,
            next_formulation=self.formulation,
        )



# COMPLETE = 'done'
CONTROL_ARM = "control_arm"
DAYS = "days"
IN_PROGRESS_APPT = "in_progress"
PARTIAL = "partial"
FILLED = "filled"
PRINT_SELECTED = "print_selected"
SOLUTION = "SOLUTION"
CANCELLED = "cancelled"
DISPENSED = "dispensed"
# SUPPOSITORY = 'SUPPOSITORY'
# SYRUP = 'SYRUP'
# TABLET = 'TABLET'
# ORAL = 'oral'
# VIAL = 'vial'
# CAPSULE = 'capsule'
# IM = 'IM'
# IV = 'IV'
# MILLIGRAM = 'mg'
# MONTHS = 'months'
# WEEKS = 'weeks'
PRESCRIPTION_ACTION = "prescription"



from django.apps import apps as django_apps

app_name = "edc_pharmacy"
DISPENSING = "DISPENSING"
PHARMACY = "PHARMACY"
PHARMACY_VIEW = "PHARMACY_VIEW"
DISPENSING_VIEW = "DISPENSING_VIEW"
PHARMACIST_ROLE = "pharmacist"
SITE_PHARMACIST_ROLE = "site_pharmacist"

pharmacy_codenames = []
for app_config in django_apps.get_app_configs():
    if app_config.name in [
        "edc_pharmacy",
    ]:
        for model_cls in app_config.get_models():
            app_name, model_name = model_cls._meta.label_lower.split(".")
            for prefix in ["add", "change", "view", "delete"]:
                if model_name == "subject" and prefix in ["add", "change", "delete"]:
                    pass
                else:
                    pharmacy_codenames.append(f"{app_name}.{prefix}_{model_name}")
pharmacy_codenames.sort()

# dispensing_models = [
#     "dispenseditem",
#     "dosageguideline",
#     "medication",
#     "prescription",
#     "prescriptionitem",
# ]
# dispensing_codenames = []
# for model_name in dispensing_models:
#     dispensing_codenames.extend([c for c in pharmacy_codenames if model_name in c])
# dispensing_models.sort()



class PrescriptionError(Exception):
    pass


class PrescriptionRefillError(Exception):
    pass


class ActivePrescriptionRefillOverlap(Exception):
    pass


class ActivePrescriptionRefillExists(Exception):
    pass


class NextPrescriptionRefillError(Exception):
    pass



from uuid import uuid4

from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from edc_pharmacy.models import MedicationStockCreateLabels
from edc_pharmacy.models.medication_stock_create_labels import Labels

from ..dispensing import Dispensing
from ..refill import create_next_refill, create_refill
from .dispensing_history import DispensingHistory


@receiver(
    post_save, sender=DispensingHistory, dispatch_uid="dispensing_history_on_post_save"
)
def dispensing_history_on_post_save(sender, instance, raw, created, **kwargs):
    if not raw:
        dispensing = Dispensing(
            rx_refill=instance.rx_refill, dispensed=instance.dispensed
        )
        instance.rx_refill.remaining = dispensing.remaining
        instance.rx_refill.save(update_fields=["remaining"])


@receiver(
    post_delete,
    sender=DispensingHistory,
    dispatch_uid="dispensing_history_on_post_delete",
)
def dispensing_history_on_post_delete(sender, instance, using=None, **kwargs):
    dispensing = Dispensing(rx_refill=instance.rx_refill, dispensed=instance.dispensed)
    instance.rx_refill.remaining = dispensing.remaining
    instance.rx_refill.save(update_fields=["remaining"])


@receiver(
    post_save,
    dispatch_uid="create_refills_on_post_save",
)
def create_refills_on_post_save(sender, instance, raw, created, **kwargs):
    if not raw:
        try:
            instance.creates_refills_from_crf
        except AttributeError:
            pass
        else:
            if instance.creates_refills_from_crf:
                create_refill(instance)
                create_next_refill(instance)


@receiver(
    post_save,
    sender=MedicationStockCreateLabels,
    dispatch_uid="create_medication_stock_labels_on_post_save",
)
def create_medication_stock_labels_on_post_save(
    sender, instance, raw, created, **kwargs
):
    if not raw:
        qty_already_created = Labels.objects.filter(
            medication_stock_create_labels=instance
        ).count()
        for i in range(0, instance.qty - qty_already_created):
            Labels.objects.create(
                medication_stock_create_labels=instance, stock_identifier=uuid4().hex
            )



from django.db import models
from edc_crf.crf_model_mixin import CrfModelMixin
from edc_model import models as edc_models
from edc_pharmacy.models import StudyMedicationCrfModelMixin
from edc_sites.models import SiteModelMixin
from edc_visit_schedule.model_mixins import OffScheduleModelMixin, OnScheduleModelMixin
from edc_visit_tracking.model_mixins import VisitModelMixin


class SubjectVisit(VisitModelMixin, edc_models.BaseUuidModel):
    class Meta(VisitModelMixin.Meta, edc_models.BaseUuidModel.Meta):
        app_label = "edc_pharmacy"


class StudyMedication(
    StudyMedicationCrfModelMixin, SiteModelMixin, edc_models.BaseUuidModel
):

    subject_visit = models.OneToOneField(SubjectVisit, on_delete=models.PROTECT)

    class Meta(CrfModelMixin.Meta, edc_models.BaseUuidModel.Meta):
        app_label = "edc_pharmacy"


class OnSchedule(OnScheduleModelMixin, edc_models.BaseUuidModel):

    pass


class OffSchedule(OffScheduleModelMixin, edc_models.BaseUuidModel):

    pass



from datetime import date, datetime
from typing import Union

import arrow
from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist

from .action_items import PrescriptionAction


def convert_to_utc_date(dte: Union[datetime, date]) -> date:
    try:
        dt = dte.date()
    except AttributeError:
        dt = arrow.get(dte).to("utc").date()
    return dt


def create_prescription(
    subject_identifier: str,
    report_datetime: datetime,
    randomizer_name: str,
    medication_name: str,
):
    medication_model_cls = django_apps.get_model("edc_pharmacy.medication")
    rx_model_cls = django_apps.get_model("edc_pharmacy.rx")

    action = PrescriptionAction(subject_identifier=subject_identifier)
    try:
        medication = medication_model_cls.objects.get(name__iexact=medication_name)
    except ObjectDoesNotExist:
        medication = medication_model_cls.objects.create(name=medication_name)
    try:
        rx_model_cls.objects.get(action_identifier=action.action_identifier)
    except ObjectDoesNotExist:
        rx_model_cls.objects.create(
            action_identifier=action.action_identifier,
            subject_identifier=subject_identifier,
            report_datetime=report_datetime,
            rx_date=report_datetime.date(),
            medication=medication,
            randomizer_name=randomizer_name,
        )



from ..models import RxRefill
from .refill_creator import RefillCreator


def create_refill(instance):
    """Creates the refill for this visit, if not already created.

    Called from signal.
    """
    if not RxRefill.objects.filter(
        rx__subject_identifier=instance.subject_visit.subject_identifier
    ).exists():
        number_of_days = 0
        if instance.subject_visit.appointment.next:
            number_of_days = (
                instance.subject_visit.appointment.next.appt_datetime
                - instance.subject_visit.appointment.appt_datetime
            ).days
        RefillCreator(
            subject_identifier=instance.subject_visit.subject_identifier,
            refill_date=instance.subject_visit.appointment.appt_datetime,
            visit_code=instance.subject_visit.appointment.visit_code,
            visit_code_sequence=instance.subject_visit.appointment.visit_code_sequence,
            number_of_days=number_of_days,
            dosage_guideline=instance.next_dosage_guideline,
            formulation=instance.next_formulation,
            make_active=True,
        )



# Generated by Django 3.2.11 on 2022-03-04 01:07

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('edc_pharmacy', '0001_initial'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='medication',
            unique_together={('name', 'display_name')},
        ),
    ]



from django.test import TestCase, tag
from edc_list_data import site_list_data
from edc_pharmacy.dosage_per_day import DosageError, dosage_per_day
from edc_pharmacy.models import (
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    Route,
    Units,
)


class TestDoseCalculator(TestCase):
    def setUp(self):
        site_list_data.initialize()
        site_list_data.autodiscover()
        medication = Medication.objects.create(
            name="flucytosine",
            display_name="Flucytosine",
        )

        Formulation.objects.create(
            medication=medication,
            strength=500,
            units=Units.objects.get(display_name__iexact="mg"),
            route=Route.objects.get(display_name__iexact="oral"),
            formulation_type=FormulationType.objects.all()[0],
        )

        medication = Medication.objects.create(
            name="flucanazole",
            display_name="Flucanazole",
        )

        Formulation.objects.create(
            medication=medication,
            strength=200,
            units=Units.objects.get(display_name__iexact="mg"),
            route=Route.objects.get(display_name__iexact="oral"),
            formulation_type=FormulationType.objects.all()[0],
        )

        medication = Medication.objects.create(
            name="ambisome",
            display_name="Ambisome",
        )

        Formulation.objects.create(
            medication=medication,
            strength=50,
            units=Units.objects.get(display_name__iexact="mg"),
            route=Route.objects.get(display_name__iexact="intravenous"),
            formulation_type=FormulationType.objects.all()[0],
        )

    @tag("14")
    def test_dosage_flucytosine(self):
        medication = Medication.objects.get(name="flucytosine")
        dosage_guideline = DosageGuideline.objects.create(
            medication=medication,
            dose_per_kg=100.0,
            dose_units=Units.objects.get(display_name__iexact="mg"),
            frequency=1.0,
            frequency_units=FrequencyUnits.objects.get(
                display_name__iexact="times per day"
            ),
        )
        self.assertEqual(
            dosage_per_day(dosage_guideline, strength=100.0, strength_units="mg"), 1.0
        )
        self.assertEqual(
            dosage_per_day(
                dosage_guideline,
                weight_in_kgs=40.0,
                strength=500.0,
                strength_units="mg",
            ),
            8.0,
        )

    def test_dosage_ambisome(self):
        medication = Medication.objects.get(name="ambisome")
        dosage_guideline = DosageGuideline.objects.create(
            medication=medication,
            dose_per_kg=10.0,
            dose_units=Units.objects.get(display_name__iexact="mg"),
            frequency=1.0,
            frequency_units=FrequencyUnits.objects.get(
                display_name__iexact="times per day"
            ),
        )
        self.assertEqual(
            dosage_per_day(dosage_guideline, strength=1.0, strength_units="mg"),
            10.0,
        )
        self.assertEqual(
            dosage_per_day(
                dosage_guideline, weight_in_kgs=40.0, strength=50.0, strength_units="mg"
            ),
            8.0,
        )

    def test_dosage_flucanazole(self):
        medication = Medication.objects.get(name="flucanazole")
        dosage_guideline = DosageGuideline.objects.create(
            medication=medication,
            dose=1200.0,
            dose_units=Units.objects.get(display_name__iexact="mg"),
            frequency=1.0,
            frequency_units=FrequencyUnits.objects.get(
                display_name__iexact="times per day"
            ),
        )
        self.assertEqual(
            dosage_per_day(dosage_guideline, strength=1.0, strength_units="mg"), 1200.0
        )
        self.assertEqual(
            dosage_per_day(dosage_guideline, strength=200, strength_units="mg"), 6.0
        )

    def test_dosage_exceptions(self):
        medication = Medication.objects.get(name="flucytosine")
        dosage_guideline = DosageGuideline.objects.create(
            medication=medication,
            dose_per_kg=100,
            dose_units=Units.objects.get(display_name__iexact="mg"),
            frequency=1,
            frequency_units=FrequencyUnits.objects.get(
                display_name__iexact="times per day"
            ),
        )

        self.assertRaises(
            DosageError,
            dosage_per_day,
            dosage_guideline,
            weight_in_kgs=40,
            strength=500,
            strength_units="kg",
        )



from django import forms

from ..models import MedicationStockReceiving


class MedicationStockReceivingForm(forms.ModelForm):
    class Meta:
        model = MedicationStockReceiving
        fields = "__all__"



from django import forms

from ..models import MedicationStockCreateLabels


class MedicationStockCreateLabelsForm(forms.ModelForm):
    class Meta:
        model = MedicationStockCreateLabels
        fields = "__all__"



from dateutil.relativedelta import relativedelta
from django.test import TestCase, tag
from edc_constants.constants import FEMALE
from edc_list_data import site_list_data
from edc_pharmacy.models import (
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    Route,
    Rx,
    Units,
)
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow


@tag("35")
class TestPrescription(TestCase):
    def setUp(self):
        site_list_data.initialize()
        site_list_data.autodiscover()
        self.subject_identifier = "12345"
        self.registered_subject = RegisteredSubject.objects.create(
            subject_identifier=self.subject_identifier,
            gender=FEMALE,
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.medication = Medication.objects.create(
            name="Flucytosine",
        )

        self.formulation = Formulation.objects.create(
            strength=500,
            units=Units.objects.get(name="mg"),
            route=Route.objects.get(display_name="Oral"),
            formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
        )

        self.dosage_guideline = DosageGuideline.objects.create(
            medication=self.medication,
            dose_per_kg=100,
            dose_units=Units.objects.get(name="mg"),
            frequency=1,
            frequency_units=FrequencyUnits.objects.get(name="day"),
        )

    def test_create_prescription(self):
        obj = Rx.objects.create(
            subject_identifier=self.subject_identifier,
            report_datetime=get_utcnow(),
            medication=self.medication,
        )
        obj.save()

    def test_verify_prescription(self):
        obj = Rx.objects.create(
            subject_identifier=self.subject_identifier,
            report_datetime=get_utcnow(),
            medication=self.medication,
        )
        obj.verified = True
        obj.verified = get_utcnow()
        obj.save()
        self.assertTrue(obj.verified)



from datetime import date, datetime
from typing import Any, Optional, Union

from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError, transaction

from ..exceptions import PrescriptionError, PrescriptionRefillError
from ..models import Rx, RxRefill
from ..utils import convert_to_utc_date


class RefillCreator:
    def __init__(
        self,
        subject_identifier: Optional[str] = None,
        refill_date: Union[datetime, date, type(None)] = None,
        visit_code: Optional[str] = None,
        visit_code_sequence: Optional[int] = None,
        number_of_days: Optional[int] = None,
        dosage_guideline: Optional[Any] = None,
        formulation: Optional[Any] = None,
        instance: Optional[Any] = None,
        make_active: Optional[bool] = None,
    ):
        if instance:
            self.subject_identifier = instance.get_subject_identifier()
            self.refill_date = convert_to_utc_date(instance.refill_date)
            self.visit_code = instance.visit_code
            self.visit_code_sequence = instance.visit_code_sequence
            self.number_of_days = instance.number_of_days
            self.dosage_guideline = instance.dosage_guideline
            self.formulation = instance.formulation

        else:
            self.subject_identifier = subject_identifier
            self.refill_date = convert_to_utc_date(refill_date)
            self.visit_code = visit_code
            self.visit_code_sequence = visit_code_sequence
            self.number_of_days = number_of_days
            self.dosage_guideline = dosage_guideline
            self.formulation = formulation
        self.make_active = True if make_active is None else make_active
        self.object = self.create()

    def create(self) -> Any:
        active = False
        if self.make_active:
            if self.active_refill:
                obj = self.active_refill
                obj.active = False
                obj.save()
            active = True
        get_opts = dict(
            rx=self.rx,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            refill_date=self.refill_date,
        )
        create_opts = dict(
            visit_code=self.visit_code,
            visit_code_sequence=self.visit_code_sequence,
            number_of_days=self.number_of_days,
            **get_opts,
        )
        try:
            obj = RxRefill.objects.get(**get_opts)
        except ObjectDoesNotExist:
            try:
                with transaction.atomic():
                    obj = RxRefill.objects.create(**create_opts)
            except IntegrityError as e:
                raise PrescriptionRefillError(e)
        else:
            obj.visit_code = self.visit_code
            obj.visit_code_sequence = self.visit_code_sequence
            obj.number_of_days = self.number_of_days
        finally:
            obj.active = active
            obj.save()
        return obj

    @property
    def rx(self) -> Any:
        """Returns Rx model instance else raises PrescriptionError"""
        opts = dict(
            subject_identifier=self.subject_identifier,
            medication=self.formulation.medication,
            rx_date__lte=self.refill_date,
        )
        try:
            obj = Rx.objects.get(**opts)
        except ObjectDoesNotExist:
            raise PrescriptionError(
                f"Subject does not have a valid prescription. Got {opts}."
            )
        else:
            if obj.rx_expiration_date and self.refill_date > obj.rx_expiration_date:
                raise PrescriptionError(
                    f"Subject prescription has expired. Got {self.subject_identifier} on {obj.rx_expiration_date}."
                )
        return obj

    @property
    def active_refill(self) -> Any:
        """Returns the 'active' RxRefill model instance or None"""
        try:
            obj = RxRefill.objects.get(rx=self.rx, active=True)
        except ObjectDoesNotExist:
            obj = None
        return obj
